/*
 * Copyright (c) 2009 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 *
 * Author(s): Torsten Hoefler <htor@cs.indiana.edu>
 *
 */

#include "nbcbench.h"
#include <unistd.h>
#ifdef HAVE_FFTW
#include <fftw3.h>
#endif


#ifdef HAVE_NBC
#ifdef HAVE_FFTW
fftw_complex *G_ARR_FFT=NULL; // array to perform FFT on!
fftw_plan G_FFTW_PLAN;
#endif

/* tests = 0 means don't do tests */
static double docompute_test(t_mpiargs *mpiargs, NBC_Handle *handle) {
  double testsum=0;

  if(mpiargs->compfunc == COMP_TIME) {
    double end;
    double time, testint, nexttest;
    
    if(mpiargs->tests) testint = mpiargs->computetime / 1e6 / mpiargs->tests;
  
    time = MPI_Wtime();
    end = time + mpiargs->computetime/1e6;
    nexttest = time + testint;
    while(time < end) {
      time = MPI_Wtime()-testsum; /* we have to calculate testsum longer */
      if(mpiargs->tests && time > nexttest && (handle != NULL)) {
        nexttest = time + testint;
        testsum -= MPI_Wtime();
        NBC_Test(handle);
        testsum += MPI_Wtime();
      }
    }
  } else if(mpiargs->compfunc == COMP_REG) {
    unsigned long testint;
    unsigned long cycle=0;
    
    if(mpiargs->tests) testint = mpiargs->computecycles / mpiargs->tests;
    if(!testint) testint = 1;
    //printf("%lu %lu %lu\n", mpiargs->tests, testint, mpiargs->computecycles);
  
    while(++cycle < mpiargs->computecycles) {
      if(mpiargs->tests && cycle % testint == 0 && (handle != NULL)) {
        testsum -= MPI_Wtime();
        NBC_Test(handle);
        testsum += MPI_Wtime();
      }
    }
  } else if(mpiargs->compfunc == COMP_FFT) {
#ifdef HAVE_FFTW
    //printf("executing FFT\n");
    fftw_execute(G_FFTW_PLAN);
#else
    printf("FFTW support not compiled in!\n");
    MPI_Abort(MPI_COMM_WORLD, 34);
#endif
  } else {
    printf("invalid computation function selected!\n");
    MPI_Abort(MPI_COMM_WORLD, 33);
  }
  return testsum;
}
#endif

static __inline__ double docompute_mpitest(t_mpiargs *mpiargs, MPI_Request *req) {
  printf("not implemented\n");
  return -1;
}

double max(double *array, int num) {
  double max;
  int i;
  
  max = array[0];
  for(i=1;i<num;i++)
    if(array[i]>max) {
      max = array[i];
    }
  return max;
}

double min(double *array, int num) {
  double min;
  int i;
  
  min = array[0];
  for(i=1;i<num;i++)
    if(array[i]<min) {
      min = array[i];
    }
  return min;
}

double avg(double *array, int num) {
  double sum;
  int i;
  
  sum = array[0];
  for(i=1;i<num;i++)
    sum += array[i];
  return sum/num;
}

double med(double *array, int num) {
  double holder;
  int x,y;
  
  /* Bubble sort method. */
  for(x = 0; x < num; x++)
   for(y = 0; y < num-1; y++)
    if(array[y] > array[y+1]) {
     holder = array[y+1];
     array[y+1] = array[y];
     array[y] = holder;
    }
  
  y = (num+1)/2;
  return array[y];
}

/* accepts string of the following formats:
 * 1. "1000-2000" : sets *min to 1000 and *max to 2000
 * 2. "1000-"     : sets *min to 1000 and leaves *max untouched
 * 3. "-2000"     : sets *max to 2000 and leaves *min untouched
 * 4. "2000"      : sets *max to 2000 and leaves *min untouched 
 */
static int readminmax(char *buf, int *min, int *max) {
  char *pos, temp[1024], *posdash;
   
  memset(temp, '\0', 1024);

  posdash=NULL;
  pos = buf;
  while(*pos != '\0') {
    if(*pos == '-') {
      /* min is before "-" */
      if(pos != buf) {
        /* if there is a min and not "-10" only a max */
        strncpy(temp, buf, pos-buf);
        *min = atoi(temp);
        memset(temp, '\0', 1024);
      }
      posdash=pos+1;
    }
    pos++;
  }
  if(posdash != NULL) {
    /* we found a dash */
    if(pos != posdash) {
      /* if there is a maximum and not "10-" only a min */
      strncpy(temp, posdash, pos-posdash);
      *max = atoi(temp);
      memset(temp, '\0', 1024);
    }
  } else {
    *max = atoi(buf);
  }

  if(*min > *max) {
    printf("min (%i) > max (%i), setting min=%i\n", *min, *max, *max);
    *min = *max;
  }

  return 0;
}

void print_help(void) {
  int rank;

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if(!rank) {
    printf("USAGE: overlap -t collop [options]\n");
    printf("options:\n");
    option_usage("-h", "print this help", NULL);
    option_usage("-a", "print times  for all nodes", NULL);
    option_usage("-v", "increase verbosity level", NULL);
    option_usage("-t", "select collective operation", "collop");
    option_usage("-s", "data size (<from>-<to>), e.g. 1-1024", "data size");
    option_usage("-p", "number of processes (<from>-<to>), e.g. 2-8", "processes");
#ifdef HAVE_NBC
#ifdef HAVE_FFTW
    option_usage("-c", "computation type, time, register, fft", "time");
#else
    option_usage("-c", "computation type, time, register", "time");
#endif
    option_usage("-i", "test interval, e.g. 1024", "bytes");
    option_usage("-x", "only for NBC_Neighbor_xchg[v] - number of neigbors, e.g., 10", "processes");
#endif
    option_usage("-o", "output file", "filename");
    option_usage("-n", "number of trials , e.g. 1000", "number");
    option_usage("-r", "select data reduction (min,max,med,avg)", "reduction");
  }

  MPI_Finalize();
  exit(0);
}

void option_usage(char *opt, char *desc, char *param) {

	if (param)
		printf("   %-2s %-12s %s\n", opt, param, desc);
	else
		printf("   %-15s %s\n", opt, desc);
}

static int read_input(int *argc, char ***argv, t_flags *flags) {
  int p;
  char *optchars = "t:vs:p:n:ahr:i:x:c:o:";
  int ret;
  extern int optind, opterr, optopt;
  extern char *optarg;
  
  MPI_Comm_size(MPI_COMM_WORLD, &p);

  /* default values for flags :) */
  flags->op = BARRIER;
  flags->verbose = 0;
  flags->printall = 0;
  flags->smin = 1<<12;
  flags->smax = 1<<18;
  flags->pmin = 1<<1;
  flags->pmax = p;
  flags->trials = 50;
  flags->red = MED;
  flags->testint = 2048;
  flags->numneighbor = 0;
  flags->compfunc = COMP_TIME;
  flags->output = NULL;
  flags->do_output = 0;


	/* get user options */
	while ( (ret = getopt(*argc, *argv, optchars)) >= 0)  {
		switch ((char)ret) {
			case '?':	/* unrecognized or badly used option */
				if (!strchr(optchars, optopt))
					continue;	/* unrecognized option */
				printf("Option %c requires an argument", optopt);
			case 't':	/* collective type */
        if(!strcmp("MPI_Allgatherv", optarg)) {
          flags->op = ALLGATHERV;
        } 
  if(!strcmp("MPI_Allgather", optarg)) {
          flags->op = ALLGATHER;
        }
  if(!strcmp("MPI_Allreduce", optarg)) {
          flags->op = ALLREDUCE;
        }
  if(!strcmp("MPI_Alltoall", optarg)) {
          flags->op = ALLTOALL;
        }
  if(!strcmp("MPI_Alltoallv", optarg)) {
          flags->op = ALLTOALLV;
        }
  if(!strcmp("MPI_Barrier", optarg)) {
          flags->op = BARRIER;
        }
  if(!strcmp("MPI_Bcast", optarg)) {
          flags->op = BCAST;
        }
  if(!strcmp("MPI_Gather", optarg)) {
          flags->op = GATHER;
        }
  if(!strcmp("MPI_Gatherv", optarg)) {
          flags->op = GATHERV;
        }
  if(!strcmp("MPI_Reduce", optarg)) {
          flags->op = REDUCE;
        }
  if(!strcmp("MPI_Reduce_scatter", optarg)) {
          flags->op = REDUCESCAT;
        }
  if(!strcmp("MPI_Scan", optarg)) {
          flags->op = SCAN;
        }
  if(!strcmp("MPI_Scatter", optarg)) {
          flags->op = SCATTER;
        }
  if(!strcmp("MPI_Scatterv", optarg)) {
          flags->op = SCATTERV;
        }

  if(!strcmp("MPI_Send", optarg)) {
          flags->op = SEND;
        }
  if(!strcmp("MPI_Neighbor_xchg", optarg)) {
          flags->op = NEIGHBOR_XCHG;
        }
  if(!strcmp("MPI_Neighbor_xchgv", optarg)) {
          flags->op = NEIGHBOR_XCHGV;
        }
				break;
      case 'v': /* verbose */
        flags->verbose = 1;
        break;
      case 's': /* data size */
        readminmax(optarg, &flags->smin, &flags->smax);
        break;
      case 'p': /* communicator size */
        readminmax(optarg, &flags->pmin, &flags->pmax);
        break;
      case 'n': /* number of trials size */
        flags->trials = atoi(optarg);
        break;
      case 'a': /* number of trials size */
        flags->printall = 1;
        break;
      case 'h': /* number of trials size */
        print_help();
        break;
      case 'i': /* test-interval */
        flags->testint = atoi(optarg);
        break;
      case 'x': /* test-interval */
        flags->numneighbor= atoi(optarg);
        break;
      case 'o': /* output file */
        flags->do_output = 1;
        if(flags->rank == 0) {
          flags->output= fopen(optarg, "w");
        }
        break;
      case 'r': /* reduction operation */
        if(!strcmp("min", optarg)) flags->red = MIN;
        if(!strcmp("max", optarg)) flags->red = MAX;
        if(!strcmp("med", optarg)) flags->red = MED;
        if(!strcmp("avg", optarg)) flags->red = AVG;
        break;
      case 'c': /* computation type */
        if(!strcmp("time", optarg)) flags->compfunc = COMP_TIME;
        if(!strcmp("register", optarg)) flags->compfunc = COMP_REG;
        if(!strcmp("fft", optarg)) flags->compfunc = COMP_FFT;
        break;
    }
  }

  /* smallest element is char */
  if(flags->smin < sizeof(char)) flags->smin=sizeof(char);
  
  /* reduction operations deal only with floats */
  if(((flags->op == ALLREDUCE) || 
     (flags->op == REDUCE) ||
     (flags->op == SCAN) ||
     (flags->op == REDUCESCAT)) && (flags->smin < sizeof(float)))
    flags->smin = sizeof(float);

  /* barrier needs only one datasize */
  if(flags->op == BARRIER) { flags->smin=1; flags->smax=1; }

  if(flags->pmax > p) flags->pmax=p;
  if(flags->pmin <= 2) flags->pmin=2;

  if(flags->trials < 10) flags->trials=10;

  /* the neighbor operations *need* a neighbor :) */
  if((   flags->op == NEIGHBOR_XCHG 
      || flags->op == NEIGHBOR_XCHGV ) 
      && flags->numneighbor == 0) flags->numneighbor=1;
  if(flags->numneighbor > p-1) flags->numneighbor=p-1;
  /* numneighbors works only with Alltoallv and neighbor exchange */
  if(   flags->numneighbor && !(
        flags->op == ALLTOALLV 
     || flags->op == NEIGHBOR_XCHG
     || flags->op == NEIGHBOR_XCHGV ))
    flags->numneighbor = 0;  

  /* we do only do 2 procs when chosen MPI_Send */
  if(flags->op == SEND) { flags->pmax=2; flags->pmin=2; }

  return 0;
}

static void print_selection(t_flags flags) {
  char collname[1024], redname[1024];

  switch(flags.op) {
    case ALLGATHERV:
      strcpy(collname, "MPI_Allgatherv");
      break; 
  case ALLGATHER:
      strcpy(collname, "MPI_Allgather");
      break;
  case ALLREDUCE:
      strcpy(collname, "MPI_Allreduce");
      break;
  case ALLTOALL:
      strcpy(collname, "MPI_Alltoall");
      break;
  case ALLTOALLV:
      strcpy(collname, "MPI_Alltoallv");
      break;
  case BARRIER:
      strcpy(collname, "MPI_Barrier");
      break;
  case BCAST:
      strcpy(collname, "MPI_Bcast");
      break;
  case GATHER:
      strcpy(collname, "MPI_Gather");
      break;
  case GATHERV:
      strcpy(collname, "MPI_Gatherv");
      break;
  case REDUCE:
      strcpy(collname, "MPI_Reduce");
      break;
  case REDUCESCAT:
      strcpy(collname, "MPI_Reduce_scatter");
      break;
  case SCAN:
      strcpy(collname, "MPI_Scan");
      break;
  case SCATTER:
      strcpy(collname, "MPI_Scatter");
      break;
  case SCATTERV:
      strcpy(collname, "MPI_Scatterv");
      break;
  case NEIGHBOR_XCHG:
      strcpy(collname, "MPI_Neighbor_xchg");
      break;
  case NEIGHBOR_XCHGV:
      strcpy(collname, "MPI_Neighbor_xchgv");
      break;

  case SEND:
      strcpy(collname, "MPI_Send");
      break;
  }

  switch(flags.red) {
    case MIN:
      strcpy(redname, "minimum");
      break;
    case MAX:
      strcpy(redname, "maximum");
      break;
    case MED:
      strcpy(redname, "median");
      break;
    case AVG:
      strcpy(redname, "average");
      break;
   }   

  printf("#------------------------------------------------------------------------\n");
  printf("#| benchmarking %s \n", collname);
  if (flags.numneighbor) printf("#| ... with %i neighbors\n", flags.numneighbor);
  printf("#| on %i to %i nodes \n", flags.pmin, flags.pmax);
  printf("#| with sizes %i bytes to %i bytes\n", flags.smin, flags.smax);
#ifdef HAVE_NBC
  if(flags.compfunc == COMP_TIME) printf("#| Using the time compute loop ");
  else if(flags.compfunc == COMP_REG) printf("#| Using fixed work in register compute loop ");
  else if(flags.compfunc == COMP_FFT) printf("#| Using FFT compute loop ");
  printf("and NBC_Test is done for each %i bytes\n", flags.testint);
#endif
  printf("#| ***** \n");
  printf("#| each measurement is repeated %i times and the %s (1) is printed ", flags.trials, redname);
  if (flags.verbose) printf("using verbose output ");
  if (flags.output) printf("and writing all to a file ");
  printf("\n");
  printf("#| 1) the reduction is done for the values of each rank and the maximum among all ranks is printed\n");
  printf("#|-----------------------------------------------------------------------\n");
  if(flags.verbose)
    printf("#p size t_{comp} (tests) :: t_{MPI} | t_{NBC} t_{NBCCOMMOV} t_{TEST} t_{WAIT} (t_{NBCCOMMOV}-t_{CALC}) (share of ov. comm) t_{NBCINIT} t_{NBCTEST}\n");
  else
    printf("#p size :: t_{MPI} | t_{NBC} (t_{NBCCOMMOV}-t_{CALC}) (share of ov. comm)\n");
  printf("#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  printf("\n");
}

int main(int argc, char **argv) {
  t_flags flags;
  int rank, tsupp;

  MPI_Init(&argc, &argv);
  /*MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &tsupp);
  if(tsupp != MPI_THREAD_MULTIPLE) {
    printf("We need MPI_THREAD_MULTIPLE!\n");
    MPI_Abort(MPI_COMM_WORLD, 0);
  }*/
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  flags.rank = rank;

  /*
  {
    char name[100];
    
    gethostname(name, 100);
    printf("process %i (rank %i) on node %s is waiting 10 secs\n", getpid(), rank, name);
    //sleep(10);
  }*/
  
  read_input(&argc, &argv, &flags);

  if(rank == 0) print_selection(flags);

  do_benchmark(flags);

  if(flags.output) fclose(flags.output);

  MPI_Finalize();
    
  return 0;
}

static void NBC_Abort( int ret, const char *string ) {
  int r;

  MPI_Comm_rank(MPI_COMM_WORLD, &r);
  printf("[NBCBENCH - %i] ERROR: %i in %s [perror: %s]\n", r, ret, string, strerror(ret*-1));
  MPI_Abort(MPI_COMM_WORLD, 1);
}


static int do_benchmark(t_flags flags) {
  t_mpiargs mpiargs;
  double t1, t2, t3;
  double time[NUM_BENCH], **trial, *time_all;
  int *ranks,size,p,rank,nprocs,tind,i,j, index, printprocs, origtrials;
  MPI_Comm comm, worldcomm;
  MPI_Group group, worldgroup;
  void *buf, *rbuf;
  char *ptr;

  worldcomm = MPI_COMM_WORLD;
  MPI_Comm_size(MPI_COMM_WORLD, &p);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_group(worldcomm,&worldgroup);

  /* allocate the 2D array */
  trial = (double**)calloc(1,sizeof(double*)*NUM_BENCH);
  for(i=0; i<NUM_BENCH;i++) {
    trial[i] = (double*)malloc(sizeof(double)*flags.trials);
    if(trial[i] == NULL) NBC_Abort(1, "malloc error");
  }

  if(p < 2) {
    printf("we need at least 2 procs (have %i)\n", p);
    exit(1);
  }

  mpiargs.sdispls = (int*)malloc(p*sizeof(int));
  if(mpiargs.sdispls == NULL) NBC_Abort(1, "malloc error");
  mpiargs.rdispls = (int*)malloc(p*sizeof(int));
  if(mpiargs.rdispls == NULL) NBC_Abort(1, "malloc error");
  mpiargs.scounts = (int*)malloc(p*sizeof(int));
  if(mpiargs.scounts== NULL) NBC_Abort(1, "malloc error");
  mpiargs.rcounts = (int*)malloc(p*sizeof(int));
  if(mpiargs.rcounts== NULL) NBC_Abort(1, "malloc error");
  mpiargs.rcountsred = (int*)malloc(p*sizeof(int));
  if(mpiargs.rcountsred== NULL) NBC_Abort(1, "malloc error");

  ranks = (int*)malloc(p * sizeof(int));
  if(ranks== NULL) NBC_Abort(1, "malloc error");
  for(i=0; i<p; i++) ranks[i] = i;

  { 
    int bufsize = 0;

    /* TODO: sort out all colls here :) */
    if(flags.op == ALLREDUCE) bufsize = flags.smax;
    else bufsize = flags.smax*p;

    buf = malloc(bufsize);
    if(buf== NULL) NBC_Abort(flags.smax*p, "malloc error (sbuf)");
    rbuf = malloc(bufsize);
    if(rbuf== NULL) NBC_Abort(flags.smax*p, "malloc error (rbuf)");
          
    for(i=0;i<bufsize;i++) {
      ptr=(char*)buf+i;
      *ptr=(char)rank;
      ptr=(char*)rbuf+i;
      *ptr=0;
    }
  }

/*  printf("buf: %p, rbuf: %p\n", buf, rbuf); */

  time_all = (double*)calloc(1,p*NUM_BENCH*sizeof(double));
  if(time_all== NULL) NBC_Abort(1, "malloc error (sbuf)");

  origtrials = flags.trials;
  /* loop over nprocs */
  for(nprocs=flags.pmin; nprocs<flags.pmax+1; nprocs=nprocs*2) {
    int res;
    /* reset trial each nprocs to orig value (might have been changed to
     * limit benchmark time */
    flags.trials = origtrials;
    /* all wait to enable overlapping comms :) */
    MPI_Barrier(worldcomm);
    /* create communicator for nprocs */
    res = MPI_Group_incl(worldgroup, nprocs, ranks, &group);
    if(res != MPI_SUCCESS) printf("error in MPI_Group_incl\n");
    res = MPI_Comm_create(worldcomm, group, &comm);
    if(res != MPI_SUCCESS) printf("error in MPI_Comm_create\n");

    /* ranks not in the comm skip this part and wait in barrier */
    if(rank >= nprocs) continue;
    
    /* create graph if necessary */
    if(flags.numneighbor) {
      int res;
      int neighbors = flags.numneighbor;
      if(neighbors > nprocs-1) {
        neighbors = nprocs-1;
        if(rank == 0) printf("#*** correcting neighbor number to %i (commsize: %i)\n", neighbors, nprocs);
      }
      mpiargs.indexes = (int*)malloc(sizeof(int)*nprocs);
      int *tmpedges= (int*)malloc(sizeof(int*)*nprocs*neighbors);
      for(i=0;i<nprocs*neighbors; i++) tmpedges[i] = MPI_PROC_NULL;
      /* go over all procs and neighbors for this proc and find a proc
       * right of it with a free neighbor slot. Fill this slot and go to
       * next free neighbor slot. Peer up from left to right. This might
       * leave to empty slots (MPI_PROC_NULL). */
      for(i=0;i<nprocs;i++) {
        int nei;
        for(nei=0;nei<neighbors;nei++) {
          int ind=i*neighbors + nei;
          /* if this has not a neighbor yet */
          if(tmpedges[ind] == MPI_PROC_NULL) {
            int found = 0,k,l;
            /* find peer process */
            for(k=i+1; k<nprocs;k++) {
              /* check if there is a connection to this process already */
              int foundme=0;
              for(l=0;l<neighbors;l++) {
                int remind = k*neighbors+l;
                if(tmpedges[remind] == i) {
                  foundme = 1;
                }
              }
              /* if there is a connection already - go to next possible peer */
              if(foundme) continue;

              /* see if there any any empty slots in peer process */
              for(l=0;l<neighbors;l++) {
                int remind = k*neighbors+l;
                if(tmpedges[remind] == MPI_PROC_NULL && !found) {
                  tmpedges[ind] = k;
                  tmpedges[remind] = i;
                  found = 1;  
      } } } } } }

      mpiargs.edges= (int*)malloc(sizeof(int*)*nprocs*neighbors);
      /* now, the adjacency list still has MPI_PROC_NULLs :-/ ....
       * filter them and arrange new array ... */
      int ind=0;
      for(i=0;i<nprocs;i++) {
        int nei;
        for(nei=0;nei<neighbors;nei++) {
          int eind=i*neighbors + nei;
          if(tmpedges[eind] != MPI_PROC_NULL) {
            mpiargs.edges[ind] = tmpedges[eind];
            ind++;
          }
        }
        mpiargs.indexes[i] = ind;
      }
      free(tmpedges);
      
      if(flags.verbose && rank == 0) {
        printf("# graph: ");
        j =0;
        for(i=0;i<nprocs*neighbors;i++) {
          if(i == mpiargs.indexes[j]) { printf(" | "); j++; }
          printf(" %i ", mpiargs.edges[i]);
        }
        printf("--");
        for(i=0;i<nprocs;i++) printf(" %i ", mpiargs.indexes[i]);
        printf("\n");
      }
      
      if(flags.op == NEIGHBOR_XCHG || flags.op == NEIGHBOR_XCHGV) {
        res = MPI_Graph_create (comm, nprocs, mpiargs.indexes, mpiargs.edges, 0, &mpiargs.graphcomm);
        if(res != MPI_SUCCESS) printf("error in MPI_Graph_create\n");
      }
    }


    /* fill mpiargs structure */
    mpiargs.comm = comm;
    mpiargs.buf = buf;
    mpiargs.sizered = size/sizeof(float);
    mpiargs.rbuf = rbuf;
    mpiargs.rank = rank;
    mpiargs.nprocs = nprocs;
#ifdef HAVE_NBC
    mpiargs.handle = (NBC_Handle*)malloc(sizeof(NBC_Handle));
#endif
    mpiargs.comm = comm;

    /* initialize synchronization stage 1 */
    sync_init_stage1(&mpiargs);

    /* loop over datasizes */
    for(size=flags.smin; size<=flags.smax; size=size*2) {
      double estmpitime=0, estnbctime=0; /* blocking estimate times for MPI and NBC */
      double  *errors = (double*)malloc(flags.trials*sizeof(double)); /* error array - if result is invalid */
      double *rerrors = (double*)malloc(flags.trials*sizeof(double)); /* error array - if result is invalid */
                                        
      if(0 == size) size=1;

      /* if we have a neighbor argument (and do not benchmark
       * NEIGHBOR_XCHGV - construct the sparse version
       * from mpiargs.indexes and mpiargs.egdes */
      if(flags.numneighbor && (flags.op != NEIGHBOR_XCHGV)) {
        int numneighbors;
        int startindex;
        int ind;
        if(rank) startindex=mpiargs.indexes[rank-1];
        else startindex = 0;
          
        for (i=0; i<nprocs; i++) {
          int neighsize=0;
          for(ind = startindex; ind<mpiargs.indexes[rank]; ind++) {
            //if(!i) printf("rank %i -> %i\n", rank, mpiargs.edges[ind]);
            if(mpiargs.edges[ind] == i) { neighsize = size; break; }
          }

          if(i) {
            mpiargs.sdispls[i] = mpiargs.sdispls[i-1]+mpiargs.scounts[i-1];
            mpiargs.rdispls[i] = mpiargs.sdispls[i-1]+mpiargs.rcounts[i-1];
          } else {
            mpiargs.sdispls[i] = neighsize;
            mpiargs.rdispls[i] = neighsize;
          }
          mpiargs.scounts[i] = neighsize;
          mpiargs.rcounts[i] = neighsize;
          mpiargs.rcountsred[i] = neighsize/sizeof(float);
        }
      } else {
        for (i=0; i<nprocs; i++) {
          mpiargs.sdispls[i] = i*size;
          mpiargs.rdispls[i] = i*size;
          mpiargs.scounts[i] = size;
          mpiargs.rcounts[i] = size;
          mpiargs.rcountsred[i] = size/sizeof(float);
      } }

      /*
      if(!rank) {
        for (i=0; i<nprocs; i++) {
          printf("rank: %i displs: %i, size: %i\n", i, mpiargs.sdispls[i], mpiargs.scounts[i]);
      } }*/


    
      /* update mpiargs */
      mpiargs.size = size;
      mpiargs.sizered = size/sizeof(float);
#ifdef HAVE_NBC
      if(flags.testint) mpiargs.tests = (size/flags.testint+1);
      else mpiargs.tests = 0;
#endif

      /*printf("[%i] warmup (size: %i)...\n", rank, mpiargs.size); */
      /* warmup */
      for(i=1; i<NBCB_WARMUP; i++) {
        mea_mpi(&flags, &mpiargs);
#ifdef HAVE_NBC
        mea_nbc(&flags, &mpiargs);
#endif
      }
      //printf("[%i] after warmup %i for size %i ...\n", rank, i, size); 

      {
        int bench=flags.trials/10;
        double *results;
        if(bench<2) bench=2;
        results=(double*)malloc(sizeof(double)*bench);

        /* try to avoid bad skew before getting the estimate */
        MPI_Barrier(comm);
        
	/* warmup */
        mea_mpi(&flags, &mpiargs);
        /* get an approximation of the benchmark time for MPI */
        for(i=0; i<bench; i++) 
          results[i] = mea_mpi(&flags, &mpiargs);
        //results[0] = max(results, bench);
        results[0] = min(results, bench);
        /* get the max of all hosts ... */
        //printf("MPI time: %lf\n", results[0]);
        MPI_Allreduce(&results[0], &results[1], 1, MPI_DOUBLE, MPI_MAX, comm);
        estmpitime = results[1];
        //if(!rank) printf("estimated MPI time: %lf\n", estmpitime);

#ifdef HAVE_NBC
        /* do the compute estimation that we compute a little bit longer
         * as we communicate with NBC */
	/* warmup */
        mea_nbc(&flags, &mpiargs);
        /* benchmark NBC */
        for(i=0; i<bench; i++) {
          results[i] = mea_nbc(&flags, &mpiargs);
        }
        results[0] = med(results, bench);
        /* get the max of all hosts ... */
        MPI_Allreduce(&results[0], &results[1], 1, MPI_DOUBLE, MPI_MAX, comm);
        estnbctime = results[1];

        mpiargs.compfunc = flags.compfunc;
        if(flags.compfunc == COMP_REG) {
          /* ramp up the computation time ... until it's NBC_COMPTIMEFACT times the estimated time for NBC */
          mpiargs.computetime = 0;
          mpiargs.computecycles = 16; /* start with 16 loops */
          while(mpiargs.computetime < estnbctime*NBC_COMPTIMEFACT) {
            //if(!flags.rank) printf("computetime < estnbctime, %lf < %lf %lu\n", mpiargs.computetime, estnbctime*NBC_COMPTIMEFACT,  mpiargs.computecycles); 
            mpiargs.computecycles *= 2;
            if(mpiargs.computecycles == 0) {
              printf("overflow of compute in rampup :-(\n");
              MPI_Finalize();
              exit(1);
            }
            /* warmup */
            docompute_test(&mpiargs, NULL);
            for(i=0; i<bench; i++) {
              usleep(1); /* this seems useless but is needed to avoid bogus timings :-( */
              results[i]  = -MPI_Wtime();
              docompute_test(&mpiargs, NULL);
              results[i] += MPI_Wtime();
              //if(!flags.rank) printf(" %lf \n", results[i]*1e6);
            }
            //if(!flags.rank) printf("\n");
            mpiargs.computetime = min(results, bench)*1e6;
          }
          //if(!rank) printf("[%i] after rampup ... %lf > %lf (=%lu cycles)\n", rank, mpiargs.computetime, estnbctime*NBC_COMPTIMEFACT, mpiargs.computecycles); 
          /* ensure that all use the same computation */
          MPI_Bcast(&mpiargs.computecycles, 1, MPI_UNSIGNED_LONG, 0, comm);
        } else if(flags.compfunc == COMP_FFT) {
          /* ramp up the FFT computation time ... until it's NBC_COMPTIMEFACT times the estimated time for NBC */
          mpiargs.computetime = 0;
          mpiargs.computefft_size = 16; /* start with size 16  */
          while(mpiargs.computetime < estnbctime*NBC_COMPTIMEFACT) {
            //if(!flags.rank) printf("computetime < estnbctime, %lf < %lf %lu\n", mpiargs.computetime, estnbctime*NBC_COMPTIMEFACT,  mpiargs.computecycles); 
            mpiargs.computefft_size *= 2;
#ifdef HAVE_FFTW
            if(G_ARR_FFT != NULL) free(G_ARR_FFT);
            G_ARR_FFT = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * mpiargs.computefft_size);
            G_FFTW_PLAN = fftw_plan_dft_1d(mpiargs.computefft_size, G_ARR_FFT, G_ARR_FFT, FFTW_FORWARD, FFTW_ESTIMATE);
#endif
            if(mpiargs.computefft_size == 0) {
              printf("overflow of FFT in rampup :-(\n");
              MPI_Finalize();
              exit(1);
            }
            /* warmup */
            docompute_test(&mpiargs, NULL);
            for(i=0; i<bench; i++) {
              usleep(1); /* this seems useless but is needed to avoid bogus timings :-( */
              results[i]  = -MPI_Wtime();
              docompute_test(&mpiargs, NULL);
              results[i] += MPI_Wtime();
              //if(!flags.rank) printf(" %lf \n", results[i]*1e6);
            }
            //if(!flags.rank) printf("\n");
            mpiargs.computetime = min(results, bench)*1e6;
          }
          //if(!rank) printf("[%i] after rampup ... %lf > %lf (=%i elements)\n", rank, mpiargs.computetime, estnbctime*NBC_COMPTIMEFACT, mpiargs.computefft_size); 
          /* ensure that all use the same computation */
          MPI_Bcast(&mpiargs.computecycles, 1, MPI_UNSIGNED_LONG, 0, comm);
          

        } else if(flags.compfunc == COMP_TIME) {
          mpiargs.computetime = estnbctime*NBC_COMPTIMEFACT;
          MPI_Bcast(&mpiargs.computetime, 1, MPI_DOUBLE, 0, comm);
        }
#endif
        free(results);
      }

      /* set estimate */
      if(estnbctime > estmpitime) mpiargs.esttime = estnbctime;
      else mpiargs.esttime = estmpitime;
      //if(!rank) printf("estimated time: %lf\n", mpiargs.esttime);

      /* ensure that each round does not take longer than 10 seconds */
      if(mpiargs.esttime * 5 /* 5 benchmarks per round */ * flags.trials > 10e6) {
        flags.trials /= 2;
        if(flags.trials < 4) flags.trials=4;
        if(flags.verbose && (rank == 0)) printf("##### longer than 10 seconds est. - new trial: %i\n", flags.trials);
      }
      MPI_Bcast(&flags.trials, 1, MPI_INT, 0, comm);

      sync_init_stage2(&mpiargs);
      //NBC_Reset_times();
      for(j=0; j<flags.trials; j++) {
        double err=0;
        
        //if(!mpiargs.rank) printf("[%i] loop #%i\n", mpiargs.rank, j);
        err += nbcb_sync(&mpiargs);

        /* normal MPI blocking */
        trial[RES_MPI][j] = mea_mpi(&flags, &mpiargs);
        //printf("[%i] took: %lf\n", mpiargs.rank, trial[RES_MPI][j]);
        
        /*
        {
          double wait;
          trial[RES_MPI][j] = -MPI_Wtime();
          wait = MPI_Wtime()+100/1e6;
          while(MPI_Wtime() < wait) {};
          trial[RES_MPI][j] += MPI_Wtime();
          trial[RES_MPI][j] *= 1e6;
        }*/

#ifdef HAVE_NBC
        err += nbcb_sync(&mpiargs);

        /* NBC blocking I... */
        trial[NBC][j] = mea_nbc(&flags, &mpiargs);

        err += nbcb_sync(&mpiargs);

        /* NBC non-blocking I... */
        trial[NBCCOMPOV][j] = mea_nbccompov(&flags, &mpiargs, &(trial[NBCCOMPTEST][j]), &(trial[NBCCOMPWAIT][j]));
        
        //err += nbcb_sync(&mpiargs);
        
        /* NBC blocking I... - testing overhead */
        trial[NBCINIT][j] = mea_nbcinit(&flags, &mpiargs, &(trial[NBCTEST][j]));
#endif

        errors[j] = err;
      }
      //if(!rank) NBC_Print_times((double)(3*flags.trials));

      { /* check if there were to many synchronization errors */
        int ii, errcnt=0;
#ifdef SYNC_WINDOW
      /* all trials benchmarked and the errors field contains the late
       * times for every trial */

        double errtime=0; /* first error-difference */
        
        /* check if a run is successful */
        MPI_Allreduce(errors, rerrors, flags.trials, MPI_DOUBLE, MPI_MAX, mpiargs.comm);
        /* count errors */
        for(ii = 0; ii<flags.trials; ii++) {
          if(rerrors[ii] > 0.0) {
            /* we found an error */
            //if(!mpiargs.rank) printf("sync error (%lf) in trial %i\n", rerrors[ii]*1e6, ii);
            errcnt++;

            /* save the first error time  - because everything else is
             * just summed up */
            if(errtime == 0.0) errtime = rerrors[ii];
          
            /* exchange with the last valid entry */
            for(j=0;j<NUM_BENCH;j++) {
              trial[j][ii] = trial[j][flags.trials-errcnt];
            }
          }
        }
        /* if we have more than 25% errors or less than 4 valid
         * measurements - double windowsize and repeat measurement */
        if((errcnt > flags.trials*0.25) || (flags.trials-errcnt < 4)) {
          size = size/2;
          mpiargs.window *= 1.5;
          //if(!mpiargs.rank) printf("# > 25%% synch errors (%i/%i) - increasing window to %lf and restarting size %i ...\n", errcnt, flags.trials, mpiargs.window*1e6, size*2);
          continue;
        }
#endif
        
        /* write data to file if requested */
        if(flags.do_output) {
          int k,t;
          /* collect all trials */
          for(t=0;t<flags.trials-errcnt;t++) {
            for(j=0;j<NUM_BENCH;j++) {
              time[j] = trial[j][t];
            }
            MPI_Gather(time, NUM_BENCH, MPI_DOUBLE, time_all, NUM_BENCH, MPI_DOUBLE, 0, comm);
            
            if(flags.output) { /* only set on rank 0 */
              for(k=0;k<p;k++) {
                fprintf(flags.output, "%i %i %i %.2lf %lu ", mpiargs.nprocs, k, mpiargs.size, mpiargs.computetime, mpiargs.tests);
                for(j=0;j<NUM_BENCH;j++) {
                  fprintf(flags.output, " %lf ", time_all[k*NUM_BENCH+j]);
                }
                fprintf(flags.output, "\n");
                fflush(flags.output);
        } } } }
        
        for(j=0;j<NUM_BENCH;j++) {
          if(flags.red == MIN) time[j] = min(trial[j], flags.trials-errcnt);
          if(flags.red == MED) time[j] = med(trial[j], flags.trials-errcnt);
          if(flags.red == MAX) time[j] = max(trial[j], flags.trials-errcnt);
          if(flags.red == AVG) time[j] = avg(trial[j], flags.trials-errcnt);
        }
      }

      MPI_Gather(time, NUM_BENCH, MPI_DOUBLE, time_all, NUM_BENCH, MPI_DOUBLE, 0, comm);

      if(rank == 0) {
#ifdef HAVE_NBC
        if(flags.verbose) {
          printf("%i %i %.2lf (%lu) ", mpiargs.nprocs, mpiargs.size, mpiargs.computetime, mpiargs.tests);
        } else {
          printf("%i %i", mpiargs.nprocs, mpiargs.size);
        }
#else
        printf("%i %i ", mpiargs.nprocs, mpiargs.size);
#endif
#ifdef SYNC_WINDOW
        if(flags.verbose) {
          printf(" [%.2lf] ", mpiargs.window*1e6);
        }
#endif
        printf(" :: ");
        if(flags.printall) {
          printprocs = nprocs;
        } else {
          double *temparr;
          int j;
          
          printprocs = 1;
          /* reduce the data from all procs to the entry of proc 0
           * (currently only maximum supported - makes sense for pseudo synch. behavior) */ 
          temparr = (double*)malloc(sizeof(double)*p);
          for(j=0; j<NUM_BENCH; j++) {
            for(i=0; i<nprocs; i++) {
              index = i*NUM_BENCH;
              temparr[i] = time_all[index+j];
            }
            time_all[j] = max(temparr, nprocs);
          }
          free(temparr);
        }
        for(i=0; i<printprocs; i++) {
          index = i*NUM_BENCH;
#ifdef HAVE_NBC
          if(flags.verbose) {
            printf(" %.2lf |", time_all[index+RES_MPI]);
            printf(" %.2lf", time_all[index+NBC]);
            printf(" %.2lf", time_all[index+NBCCOMPOV]);
            printf(" %.2lf", time_all[index+NBCCOMPTEST]);
            printf(" %.2lf", time_all[index+NBCCOMPWAIT]);
            printf(" ( %.2lf )", time_all[index+NBCCOMPOV]-mpiargs.computetime);
            printf(" ( %.2lf / %.2lf )", 1-(time_all[index+NBCCOMPOV]-mpiargs.computetime)/time_all[index+NBC], 1-(time_all[index+NBCCOMPOV]-mpiargs.computetime)/time_all[index+RES_MPI]);
            printf(" %.2lf", time_all[index+NBCINIT]);
            printf(" %.2lf", time_all[index+NBCTEST]);
            printf(" :: ");
          } else {
            printf(" %.2lf |", time_all[index+RES_MPI]);
            printf(" %.2lf", time_all[index+NBC]);
            printf(" ( %.2lf )", time_all[index+NBCCOMPOV]-mpiargs.computetime);
            printf(" ( %.2lf )", 1-(time_all[index+NBCCOMPOV]-mpiargs.computetime)/time_all[index+NBC]);
            printf(" :: ");
          }
#endif
#ifndef HAVE_NBC
          printf(" %.2lf ::", time_all[index+RES_MPI]);
#endif
        }
      }
      if (rank==0) printf("\n");
      free(errors);
      free(rerrors);
    }
    if (rank==0) printf("\n");
    
    if(flags.numneighbor) {
      free(mpiargs.indexes);
      free(mpiargs.edges);
    }
  }

  return 1;
}


static __inline__ double mea_mpi(t_flags *flags, t_mpiargs *mpiargs) {
  double t1, t2;
  double t3;

  switch(flags->op) {
    case ALLGATHERV:
      t1 = MPI_Wtime();
      MPI_Allgatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm);
      t2 = MPI_Wtime();
      break;
 
  case ALLGATHER:
      t1 = MPI_Wtime();
      MPI_Allgather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case ALLREDUCE:
      t1 = MPI_Wtime();
      MPI_Allreduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case ALLTOALL:
      t1 = MPI_Wtime();
      MPI_Alltoall (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case ALLTOALLV:
      t1 = MPI_Wtime();
      MPI_Alltoallv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case BARRIER:
      t1 = MPI_Wtime();
      MPI_Barrier (mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case BCAST:
      t1 = MPI_Wtime();
      MPI_Bcast (mpiargs->buf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case GATHER:
      t1 = MPI_Wtime();
      MPI_Gather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case GATHERV:
      t1 = MPI_Wtime();
      MPI_Gatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, 0, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case REDUCE:
      t1 = MPI_Wtime();
      MPI_Reduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_INT, MPI_SUM, 0, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case REDUCESCAT:
      t1 = MPI_Wtime();
      MPI_Reduce_scatter (mpiargs->buf, mpiargs->rbuf, mpiargs->rcountsred, MPI_FLOAT, MPI_SUM, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case SCAN:
      t1 = MPI_Wtime();
      MPI_Scan (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case SCATTER:
      t1 = MPI_Wtime();
      MPI_Scatter (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

  case SCATTERV:
      t1 = MPI_Wtime();
      MPI_Scatterv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->size*mpiargs->nprocs, MPI_BYTE, 0, mpiargs->comm);
      t2 = MPI_Wtime();
      break;

#ifdef HAVE_NBC
  case NEIGHBOR_XCHG:
      {
        NBC_Handle handle;
        t1 = MPI_Wtime();
        NBC_Ineighbor_xchg (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->graphcomm, &handle);
        NBC_Wait(&handle);
        t2 = MPI_Wtime();
      }
      break;

  case NEIGHBOR_XCHGV:
      {
        NBC_Handle handle;
        t1 = MPI_Wtime();
        NBC_Ineighbor_xchgv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->graphcomm, &handle);
        NBC_Wait(&handle);
        t2 = MPI_Wtime();
      }
      break;
#endif

  case SEND:
      { 
        if(mpiargs->rank == 0) {
          t1 = MPI_Wtime();
          MPI_Send(mpiargs->buf, mpiargs->size, MPI_BYTE, 1, 1, mpiargs->comm);
          t2 = MPI_Wtime();
        } else if(mpiargs->rank == 1) {
          t1 = MPI_Wtime();
          MPI_Recv(mpiargs->buf, mpiargs->size, MPI_BYTE, 0, 1, mpiargs->comm, MPI_STATUS_IGNORE);
          t2 = MPI_Wtime();
        }
      }
      break;
  default:
    printf("operation %i not supported\n", flags->op);
  }
  t2=t2-t1;
  t3 = t2*1e6;

  return t3;
}

static __inline__ double mea_nbccomp(t_flags *flags, t_mpiargs *mpiargs) {
  double t1, t2;
  double t3=0;

#ifdef HAVE_NBC
  switch(flags->op) {
    case ALLGATHERV:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Iallgatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;
 
  case ALLGATHER:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Iallgather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case ALLREDUCE:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Iallreduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case ALLTOALL:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ialltoall (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case ALLTOALLV:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ialltoallv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case BARRIER:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ibarrier (mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case BCAST:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ibcast (mpiargs->buf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case GATHER:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Igather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case GATHERV:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Igatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case REDUCE:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ireduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_INT, MPI_SUM, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case REDUCESCAT:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ireduce_scatter (mpiargs->buf, mpiargs->rbuf, mpiargs->rcountsred, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case SCAN:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Iscan (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case SCATTER:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Iscatter (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case SCATTERV:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Iscatterv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->size*mpiargs->nprocs, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case NEIGHBOR_XCHG:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ineighbor_xchg (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case NEIGHBOR_XCHGV:
      t1 = MPI_Wtime();
      docompute_test(mpiargs, NULL);
      NBC_Ineighbor_xchgv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case SEND:
      { 
        MPI_Request req;
        if(mpiargs->rank == 0) {
          t1 = MPI_Wtime();
          docompute_mpitest(mpiargs, NULL);
          MPI_Isend(mpiargs->buf, mpiargs->size, MPI_BYTE, 1, 1, mpiargs->comm, &req);
          MPI_Wait(&req, MPI_STATUS_IGNORE);
          t2 = MPI_Wtime();
        } else if(mpiargs->rank == 1) {
          t1 = MPI_Wtime();
          docompute_mpitest(mpiargs, NULL);
          MPI_Irecv(mpiargs->buf, mpiargs->size, MPI_BYTE, 0, 1, mpiargs->comm, &req);
          MPI_Wait(&req, MPI_STATUS_IGNORE);
          t2 = MPI_Wtime();
        }
      }
      break;
  }
  t2=t2-t1;
  t3 = t2*1e6;
#endif

  return t3;
}

static __inline__ double mea_nbc(t_flags *flags, t_mpiargs *mpiargs) {
  double t1, t2;
  double t3=0;

#ifdef HAVE_NBC
  switch(flags->op) {
    case ALLGATHERV:
      t1 = MPI_Wtime();
      NBC_Iallgatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;
 
  case ALLGATHER:
      t1 = MPI_Wtime();
      NBC_Iallgather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case ALLREDUCE:
      t1 = MPI_Wtime();
      NBC_Iallreduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case ALLTOALL:
      t1 = MPI_Wtime();
      NBC_Ialltoall (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case ALLTOALLV:
      t1 = MPI_Wtime();
      NBC_Ialltoallv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case BARRIER:
      t1 = MPI_Wtime();
      NBC_Ibarrier (mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case BCAST:
      t1 = MPI_Wtime();
      NBC_Ibcast (mpiargs->buf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case GATHER:
      t1 = MPI_Wtime();
      NBC_Igather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case GATHERV:
      t1 = MPI_Wtime();
      NBC_Igatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case REDUCE:
      t1 = MPI_Wtime();
      NBC_Ireduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_INT, MPI_SUM, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case REDUCESCAT:
      t1 = MPI_Wtime();
      NBC_Ireduce_scatter (mpiargs->buf, mpiargs->rbuf, mpiargs->rcountsred, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case SCAN:
      t1 = MPI_Wtime();
      NBC_Iscan (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case SCATTER:
      t1 = MPI_Wtime();
      NBC_Iscatter (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case SCATTERV:
      t1 = MPI_Wtime();
      NBC_Iscatterv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->size*mpiargs->nprocs, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case NEIGHBOR_XCHG:
      t1 = MPI_Wtime();
      NBC_Ineighbor_xchg (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;

  case NEIGHBOR_XCHGV:
      t1 = MPI_Wtime();
      NBC_Ineighbor_xchgv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      NBC_Wait(mpiargs->handle);
      t2 = MPI_Wtime();
      break;



  case SEND:
      { 
        MPI_Request req;
        if(mpiargs->rank == 0) {
          t1 = MPI_Wtime();
          MPI_Isend(mpiargs->buf, mpiargs->size, MPI_BYTE, 1, 1, mpiargs->comm, &req);
          MPI_Wait(&req, MPI_STATUS_IGNORE);
          t2 = MPI_Wtime();
        } else if(mpiargs->rank == 1) {
          t1 = MPI_Wtime();
          MPI_Irecv(mpiargs->buf, mpiargs->size, MPI_BYTE, 0, 1, mpiargs->comm, &req);
          MPI_Wait(&req, MPI_STATUS_IGNORE);
          t2 = MPI_Wtime();
        }
      }
      break;
  }
  t2 -= t1;
  t3  = t2*1e6;
#endif

  return t3;
}

static __inline__ double mea_nbccompov(t_flags *flags, t_mpiargs *mpiargs, double *testtime, double *waittime) {
  double t1, t2, t3, t4;
  double t=0;

#ifdef HAVE_NBC
  switch(flags->op) {
    case ALLGATHERV:
      t1 = MPI_Wtime();
      NBC_Iallgatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;
 
  case ALLGATHER:
      t1 = MPI_Wtime();
      NBC_Iallgather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case ALLREDUCE:
      t1 = MPI_Wtime();
      NBC_Iallreduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case ALLTOALL:
      t1 = MPI_Wtime();
      NBC_Ialltoall (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case ALLTOALLV:
      t1 = MPI_Wtime();
      NBC_Ialltoallv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case BARRIER:
      t1 = MPI_Wtime();
      NBC_Ibarrier (mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case BCAST:
      t1 = MPI_Wtime();
      NBC_Ibcast (mpiargs->buf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case GATHER:
      t1 = MPI_Wtime();
      NBC_Igather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case GATHERV:
      t1 = MPI_Wtime();
      NBC_Igatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case REDUCE:
      t1 = MPI_Wtime();
      NBC_Ireduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_INT, MPI_SUM, 0, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case REDUCESCAT:
      t1 = MPI_Wtime();
      NBC_Ireduce_scatter (mpiargs->buf, mpiargs->rbuf, mpiargs->rcountsred, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case SCAN:
      t1 = MPI_Wtime();
      NBC_Iscan (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case SCATTER:
      t1 = MPI_Wtime();
      NBC_Iscatter (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case SCATTERV:
      t1 = MPI_Wtime();
      NBC_Iscatterv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->size*mpiargs->nprocs, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case NEIGHBOR_XCHG:
      t1 = MPI_Wtime();
      NBC_Ineighbor_xchg (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;

  case NEIGHBOR_XCHGV:
      t1 = MPI_Wtime();
      NBC_Ineighbor_xchgv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      t4 = docompute_test(mpiargs, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      t3 = MPI_Wtime();
      break;


  case SEND:
      { 
        MPI_Request req;
        if(mpiargs->rank == 0) {
          t1 = MPI_Wtime();
          MPI_Isend(mpiargs->buf, mpiargs->size, MPI_BYTE, 1, 1, mpiargs->comm, &req);
          t4 = docompute_mpitest(mpiargs, &req);
          t2 = MPI_Wtime();
          MPI_Wait(&req, MPI_STATUS_IGNORE);
          t3 = MPI_Wtime();
        } else if(mpiargs->rank == 1) {
          t1 = MPI_Wtime();
          MPI_Irecv(mpiargs->buf, mpiargs->size, MPI_BYTE, 0, 1, mpiargs->comm, &req);
          t4 = docompute_mpitest(mpiargs, &req);
          t2 = MPI_Wtime();
          MPI_Wait(&req, MPI_STATUS_IGNORE);
          t3 = MPI_Wtime();
        }
      }
      break;
  }
  t2=t3-t2;
  *waittime = t2*1e6;
  t3=t3-t1;
  t = t3*1e6;
  *testtime = t4*1e6;
#endif

  return t;
}

static __inline__ double mea_nbcinit(t_flags *flags, t_mpiargs *mpiargs, double *testtime) {
  double t1, t2, t3;
  double t4=0;

#ifdef HAVE_NBC
  switch(flags->op) {
    case ALLGATHERV:
      t1 = MPI_Wtime();
      NBC_Iallgatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;
 
  case ALLGATHER:
      t1 = MPI_Wtime();
      NBC_Iallgather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case ALLREDUCE:
      t1 = MPI_Wtime();
      NBC_Iallreduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case ALLTOALL:
      t1 = MPI_Wtime();
      NBC_Ialltoall (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case ALLTOALLV:
      t1 = MPI_Wtime();
      NBC_Ialltoallv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case BARRIER:
      t1 = MPI_Wtime();
      NBC_Ibarrier (mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case BCAST:
      t1 = MPI_Wtime();
      NBC_Ibcast (mpiargs->buf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case GATHER:
      t1 = MPI_Wtime();
      NBC_Igather (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case GATHERV:
      t1 = MPI_Wtime();
      NBC_Igatherv (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->rcounts, mpiargs->rdispls, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case REDUCE:
      t1 = MPI_Wtime();
      NBC_Ireduce (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_INT, MPI_SUM, 0, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case REDUCESCAT:
      t1 = MPI_Wtime();
      NBC_Ireduce_scatter (mpiargs->buf, mpiargs->rbuf, mpiargs->rcountsred, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case SCAN:
      t1 = MPI_Wtime();
      NBC_Iscan (mpiargs->buf, mpiargs->rbuf, mpiargs->sizered, MPI_FLOAT, MPI_SUM, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case SCATTER:
      t1 = MPI_Wtime();
      NBC_Iscatter (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case SCATTERV:
      t1 = MPI_Wtime();
      NBC_Iscatterv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->size*mpiargs->nprocs, MPI_BYTE, 0, mpiargs->comm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case NEIGHBOR_XCHG:
      t1 = MPI_Wtime();
      NBC_Ineighbor_xchg (mpiargs->buf, mpiargs->size, MPI_BYTE, mpiargs->rbuf, mpiargs->size, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;

  case NEIGHBOR_XCHGV:
      t1 = MPI_Wtime();
      NBC_Ineighbor_xchgv (mpiargs->buf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->rbuf, mpiargs->scounts, mpiargs->sdispls, MPI_BYTE, mpiargs->graphcomm, mpiargs->handle);
      t2 = MPI_Wtime();
      NBC_Test(mpiargs->handle);
      t3 = MPI_Wtime();
      NBC_Wait(mpiargs->handle);
      break;


  case SEND:
      { 
        MPI_Request req;
        int flag;
        if(mpiargs->rank == 0) {
          t1 = MPI_Wtime();
          MPI_Isend(mpiargs->buf, mpiargs->size, MPI_BYTE, 1, 1, mpiargs->comm, &req);
          t2 = MPI_Wtime();
          MPI_Test(&req, &flag, MPI_STATUS_IGNORE);
          t3 = MPI_Wtime();
          MPI_Wait(&req, MPI_STATUS_IGNORE);
        } else if(mpiargs->rank == 1) {
          t1 = MPI_Wtime();
          MPI_Irecv(mpiargs->buf, mpiargs->size, MPI_BYTE, 0, 1, mpiargs->comm, &req);
          t2 = MPI_Wtime();
          MPI_Test(&req, &flag, MPI_STATUS_IGNORE);
          t3 = MPI_Wtime();
        }
      }
      break;
  }
  t3=t3-t2;
  t2=t2-t1;
  t4 = t2*1e6;
  *testtime = t3*1e6;
#endif
  return t4;
}
